import time

from pygsmmodule.shared.support import SimSupportFunctions
from pygsmmodule.shared.typesandconsts import PyGsmModuleTimeoutError, SimSpecialConsts


class BaseUart(object):
    def __init__(self, serial):
        self._serial = serial

    def _send_raw_bytes(self, data, timeout=1000):
        bytes_to_send = len(data)
        sent_bytes = 0
        start = time.time()

        while sent_bytes < bytes_to_send:
            if SimSupportFunctions.time_delta(start) >= timeout:
                raise PyGsmModuleTimeoutError

            sent_bytes += self._serial.write(data[sent_bytes:])
            if sent_bytes == 0:
                time.sleep(0.001)
                continue

        return SimSupportFunctions.time_delta(start)

    def writeln(self, data, timeout=1000):
        if type(data) == str:
            data = str(data).encode("ascii")

        data = bytearray(data) + bytearray([SimSpecialConsts.cr, SimSpecialConsts.lf])

        return self._send_raw_bytes(data, timeout)

    def send_simple_request(self, command, timeout=1000, possible_results = None):
        if type(command) == str:
            command = str(command).encode("ascii")

        if possible_results is None:
            possible_results = ["OK", "ERROR"]

        self._serial.flush()
        spent_time = self.writeln(command, timeout)
        timeout -= spent_time

        return self._read_simple_result(timeout, possible_results)

    def _read_simple_result(self, timeout, possible_results):
        tm_begin = time.time()

        read_bytes_qty = 0
        buffer = bytearray()
        while True:
            if SimSupportFunctions.time_delta(tm_begin) >= timeout:
                raise PyGsmModuleTimeoutError

            while True:
                b = self._serial.read(100)

                if (b is not None) and (len(b) >= 1):
                    buffer += bytearray(b)

                    read_bytes_qty += len(b)
                    continue
                else:
                    break

            if read_bytes_qty == 0:
                time.sleep(0.005)
                continue

            strings = SimSupportFunctions.parse_data(buffer[:])

            if len(strings) == 0:
                time.sleep(0.01)
                continue

            last_string = SimSupportFunctions.get_last_non_empty_string(strings[:])

            if last_string in possible_results:
                return SimSupportFunctions.convert_to_string(SimSupportFunctions.remove_end_result(strings))

        return None



import time


class SimSupportFunctions:
    @staticmethod
    def time_delta(tm_begin):
        return (time.time() - tm_begin) * 1000.0

    @staticmethod
    def parse_data(data, separator="\r", encoding="ascii"):
        if type(data) == bytearray:
            data = data.decode(encoding)

        return data.split(separator)

    @staticmethod
    def get_last_non_empty_string(data):
        for item in reversed(data):
            if len(str(item).strip()) > 0:
                return str(item).strip()

        return None

    @staticmethod
    def remove_end_result(data):
        items_qty = len(data)
        found_index = None

        for index, value in enumerate(reversed(data)):
            if str(value).strip():
                found_index = index
                break

        return data[:items_qty - (found_index + 1)] if found_index is not None else None

    @staticmethod
    def convert_to_string(values, strip=False):
        return "".join(values) if not strip else ("".join(values)).strip()



class SimSpecialConsts:
    ctrl_z = 26    # ascii character for ctr+z. End of a SMS.
    cr = 0x0d  # ascii character for carriage return.
    lf = 0x0a  # ascii character for line feed.


class Error(Exception):
    pass


class PyGsmModuleTimeoutError(Error):
    pass



from .imei.imei import ImeiSupport
from .shared.support import SimSupportFunctions
from .shared.typesandconsts import Error, SimSpecialConsts, PyGsmModuleTimeoutError
from .supplementary.get_imei import GetImei
from .uart_support.base_uart import BaseUart

__version__ = "0.1.1.0"



import random
"""
IMEI validation (CRC check) and fake IMEI generation.

You can find more documentation about IMEI here:

http://en.wikipedia.org/wiki/International_Mobile_Station_Equipment_Identity
"""


class ImeiSupport:
    """
    Class for IMEI validation and fake IMEIs generation (by known IMEI). Can check that 15-digit IMEI (std.
    length of IMEI) is valid. Or can generate needful amount of fake IMEI by any known IMEI.
    """

    IMEI_LENGTH = 15

    @staticmethod
    def next(imei):
        """
        Generates next IMEI for the given IMEI. So, using this method you can can generate needful amount of fake IMEIs
        (for example, for testing purpose).

        :param imei: string or integer representation of 15-digit IMEI (std. length of IMEI)

        :return: next value for IMEI
        """
        imei = int(imei)

        imei += 10

        imei //= 10
        check = ImeiSupport.checksum(imei)

        return imei * 10 + check

    @staticmethod
    def checksum(imei):
        """
        Generates check sum for IMEI

        :param imei: string or integer representation of 15-digit IMEI or 14-digits IMEI without CRC.

        :return: checksum for IMEI. If you entered 15-digit (std. IMEI) and IMEI is correct it's will be the same as last
        15-th digit
        """
        imei = int(imei)

        if len(str(imei)) == ImeiSupport.IMEI_LENGTH:
            imei //= 10

        sum = [0] * (ImeiSupport.IMEI_LENGTH - 1)

        mod = 10
        for i in range(1, ImeiSupport.IMEI_LENGTH):
            index = i - 1
            sum[index] = int(imei % mod)

            if i % 2 != 0:
                sum[index] *= 2

            if sum[index] >= 10:
                sum[index] = int(sum[index] % 10 + (sum[index] / 10))

            imei /= mod

        check = 0

        for i in range(len(sum)):
            check += sum[i]

        return (check * 9) % 10

    @staticmethod
    def is_valid(imei):
        """
        Check that IMEI is valid.

        :param imei: string or integer representation of 15-digit IMEI (std. IMEI)

        :return: True when IMEI is correct, False otherwise
        """

        imei = int(imei)
        crc = ImeiSupport.checksum(imei // 10)

        return crc == (imei % 10)

    @staticmethod
    def generate_new():
        """
        Generates random IMEI number

        :return: fake IMEI number
        """
        imei = 0

        qty = ImeiSupport.IMEI_LENGTH
        while qty > 0:

            if qty != ImeiSupport.IMEI_LENGTH:
                imei = imei * 10 + random.randint(0, 9)
            else:
                imei = random.randint(1, 9)

            qty -= 1

        crc = ImeiSupport.checksum(imei)

        imei = imei * 10 + crc

        assert ImeiSupport.is_valid(imei) == True
        assert ImeiSupport.is_valid(str(imei)) == True

        return str(imei)



from pygsmmodule.uart_support.base_uart import BaseUart


class GetImei(object):
    def __init__(self):
        pass

    @staticmethod
    def get_imei(serial, timeout=1000):
        communicator = BaseUart(serial)

        res = communicator.send_simple_request("AT+GSN", timeout)
        return str(res).strip() if res else res



import os
import re

from pip.req import parse_requirements

from pygsmmodule import __version__

from setuptools import find_packages, setup


install_reqs = parse_requirements("requirements.txt", session=False)
requirements = [str(ir.req) for ir in install_reqs]
package_name = "pygsmmodule"
hyphen_package_name = package_name.replace("_", "-")


def read_version():
    regexp = re.compile(r"^__version__\s*=\s*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__), package_name, "__init__.py")
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError("Cannot find version in {}".format(init_py))


if __name__ == "__main__":
    packages_to_remove = ["script", "tests"]
    packages = find_packages()

    for item in packages_to_remove:
        if item in packages:
            packages.remove(item)

    setup(
        name="pygsmmodule",
        # packages=["pygsmmodule"],  # this must be the same as the name above
        packages=packages,
        version=__version__,
        description="GSM modem control libraty",
        author="Bohdan Danishevsky",
        author_email="dbn@aminis.com.ua",
        url="https://github.com/JFF-Bohdan/{}".format(package_name),  # use the URL to the github repo
        keywords=["GSM", "sim module", "SIM-800", "SIM 800", " SIM-900", "SIM 900"],  # arbitrary keywords
        install_requires=requirements,
        classifiers=[],
        license="MIT",
        zip_safe=False
    )
